use binrw::BinReaderExt;

mod newtypes;

mod file;
pub mod map;
pub mod bot;
pub mod demo;

#[derive(Debug)]
pub enum Error{
	UnexpectedFourCC,
	Header(file::Error),
	Map(map::Error),
	Bot(bot::Error),
	Demo(demo::Error),
}
impl std::fmt::Display for Error{
	fn fmt(&self,f:&mut std::fmt::Formatter<'_>)->std::fmt::Result{
		write!(f,"{self:?}")
	}
}
impl std::error::Error for Error{}

pub enum SNF<R:BinReaderExt>{
	Map(map::StreamableMap<R>),
	Bot(bot::StreamableBot<R>),
	Demo(demo::StreamableDemo<R>),
}

pub fn read_snf<R:BinReaderExt>(input:R)->Result<SNF<R>,Error>{
	let file=file::File::new(input).map_err(Error::Header)?;
	Ok(match file.fourcc(){
		file::FourCC::Map=>SNF::Map(map::StreamableMap::new(file).map_err(Error::Map)?),
		file::FourCC::Bot=>SNF::Bot(bot::StreamableBot::new(file).map_err(Error::Bot)?),
		file::FourCC::Demo=>SNF::Demo(demo::StreamableDemo::new(file).map_err(Error::Demo)?),
	})
}
pub fn read_map<R:BinReaderExt>(input:R)->Result<map::StreamableMap<R>,Error>{
	let file=file::File::new(input).map_err(Error::Header)?;
	match file.fourcc(){
		file::FourCC::Map=>Ok(map::StreamableMap::new(file).map_err(Error::Map)?),
		_=>Err(Error::UnexpectedFourCC)
	}
}
pub fn read_bot<R:BinReaderExt>(input:R)->Result<bot::StreamableBot<R>,Error>{
	let file=file::File::new(input).map_err(Error::Header)?;
	match file.fourcc(){
		file::FourCC::Bot=>Ok(bot::StreamableBot::new(file).map_err(Error::Bot)?),
		_=>Err(Error::UnexpectedFourCC)
	}
}
pub fn read_demo<R:BinReaderExt>(input:R)->Result<demo::StreamableDemo<R>,Error>{
	let file=file::File::new(input).map_err(Error::Header)?;
	match file.fourcc(){
		file::FourCC::Demo=>Ok(demo::StreamableDemo::new(file).map_err(Error::Demo)?),
		_=>Err(Error::UnexpectedFourCC)
	}
}

#[cfg(test)]
mod tests {
	//use super::*;

	#[test]
	fn test() {
	}
}
